/* FILE: viotoomf.cc                 -*-Mode: c++-*-
 *
 * Program to convert from vio to omf vector field file formats.
 *
 */

#include <stdio.h>
#include "machine.h"
#include "misc.h"
#include "template.h"
#include "vector.h"
#include "vecfile.h"

enum OutputMode { OM_ASCII=1, OM_BINARY_4=2, OM_BINARY_8=3 };

void Usage(void)
{
  fprintf(stderr,"Usage: viotoomf [-b4|-b8] "
	  "[input.vio |<input.vio] >output.omf\n");
  exit(1);
}

void DumpFloat4(FILE* outfile,REAL4 x)
{
  unsigned char *buf=(unsigned char *)(&x);
#if (BYTEORDER == 4321)
  swap_buf(buf,sizeof(REAL4),1); // Convert from LSB to MSB order
#endif
  fwrite((char *)buf,sizeof(REAL4),1,outfile);
}

void DumpFloat8(FILE* outfile,REAL8 x)
{
  unsigned char *buf=(unsigned char *)(&x);
#if (BYTEORDER == 4321)
  swap_buf(buf,sizeof(REAL8),1); // Convert from LSB to MSB order
#endif
  fwrite((char *)buf,sizeof(REAL8),1,outfile);
}

void ProcessArgs(int argc,char **argv,char* &inname,FILE* &infile,
		 OutputMode &om)
{
  if(argc>3) Usage();

  // Set defaults
  inname=NULL;
  om=OM_ASCII;

  // Loop through arguments
  for(int i=1;i<argc;i++) {
    if(argv[i][0]!='-') { // Input file
      if(inname!=NULL) Usage(); // Error: we already opened an input file
      inname=argv[i];
    }
    else { // Option
      if(strcmp(argv[i],"-b4")==0)      om=OM_BINARY_4;
      else if(strcmp(argv[i],"-b8")==0) om=OM_BINARY_8;
      else Usage();
    }
  }

  // Open input file
  if(inname==NULL) infile=stdin;
  else {
    infile=fopen(inname,"rb"); // Vio files are always binary
    if(infile==NULL) {
      fprintf(stderr,"ERROR: Unable to open input file %s\n",inname);
      Usage();
    }
  }
}



int main(int argc,char **argv)
{
  FILE *infile;
  OutputMode om;
  char *inname,*omfname;

  ProcessArgs(argc,argv,inname,infile,om);

  if(inname!=NULL) {
    omfname=new char[strlen(inname)+1+4];
    strcpy(omfname,inname);
    char *cptr=strstr(omfname,".vio");
    if(cptr!=NULL) strcpy(cptr,".omf");
    else           strcat(omfname,".omf");
  }
  else {
    inname="stdin";
    omfname=new char[strlen("unknown.omf")+1];
    strcpy(omfname,"unknown.omf");
  }

  // Open and read vio file
  VioFileInput vfi(inname,infile,NULL,NULL);
  GridVec3f *grid=vfi.NewRectGrid();

  INT4m xdim,ydim,zdim;
  grid->GetSize(xdim,ydim,zdim);

  if(zdim!=1) {
    fprintf(stderr,"Illegal z dimension in input file: %d (should be 1)\n",
	    zdim);
    exit(1);
  }

  // Write omf file
  fprintf(stdout,"# OOMMF: rectangular mesh v0.99\n");
  fprintf(stdout,"# Segment count: 1\n");
  fprintf(stdout,"# Begin: Segment\n");
  fprintf(stdout,"# Begin: Header\n");
  fprintf(stdout,"# Title: %s\n",omfname);
  fprintf(stdout,"# Desc: Omf file created from vio file by viotoomf\n");
  fprintf(stdout,"# meshunit: unknown\n");
  fprintf(stdout,"# xbase: 0\n# ybase: 0\n# zbase: 0\n");
  fprintf(stdout,"# xstepsize: 1\n# ystepsize: 1\n# zstepsize: 1\n");
  fprintf(stdout,"# xnodes: %d\n",xdim);
  fprintf(stdout,"# ynodes: %d\n",ydim);
  fprintf(stdout,"# znodes: 1\n");
  fprintf(stdout,"# xmin: -0.5\n# ymin: -0.5\n# zmin: -0.5\n");
  fprintf(stdout,"# xmax: %.1f\n# ymax: %.1f\n# zmax:  0.5\n",
	  double(xdim)-0.5,double(ydim)-0.5);
  fprintf(stdout,"# valueunit: Ms\n");
  fprintf(stdout,"# valuemultiplier: 1\n");
  fprintf(stdout,"# ValueRangeMaxMag:  1.\n");
  fprintf(stdout,"# ValueRangeMinMag:  1e-8\n");
  fprintf(stdout,"# End: Header\n");

  // Print grid data
  if(om==OM_BINARY_4) { // 4-byte binary output
    fprintf(stdout,"# Begin: data binary 4\n");
    DumpFloat4(stdout,1234567.);
    INT4m i,j,k;
    Vec3f v;
    for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
      v=grid->GridVec(i,j,k);
      DumpFloat4(stdout,v.x);
      DumpFloat4(stdout,v.y);
      DumpFloat4(stdout,v.z);
    }
    fprintf(stdout,"\n# End: data binary 4\n");
  }
  else if(om==OM_BINARY_8) { // 8-byte binary output
    fprintf(stdout,"# Begin: data binary 8\n");
    DumpFloat8(stdout,123456789012345.);
    INT4m i,j,k;
    Vec3f v;
    for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
      v=grid->GridVec(i,j,k);
      DumpFloat8(stdout,v.x);
      DumpFloat8(stdout,v.y);
      DumpFloat8(stdout,v.z);
    }
    fprintf(stdout,"\n# End: data binary 8\n");
  }
  else { // ASCII output
    fprintf(stdout,"# Begin: data text\n");
    INT4m i,j,k;
    Vec3f v;
    for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
      v=grid->GridVec(i,j,k);
      fprintf(stdout,"% -14.8g % -14.8g % -14.8g\n",v.x,v.y,v.z);
    }
    fprintf(stdout,"# End: data text\n");
  }
  fprintf(stdout,"# End: segment\n");


  // Clean up
  fflush(stdout);
  delete grid;
  delete[] omfname;
  return 0;
}
