/* FILE: maginit.cc	      -*-Mode: c++-*-
 *
 * This file contains function definitions for the mag element
 *  initialization routines.
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "misc.h"
#include "vector.h"
#include "vecio.h"
#include "magelt.h"
#include "maginit.h"
#include "grid.h"  // Only needed for PERTURBATION_SIZE

/* End includes */

// Uncomment the next block to disable random perturbations in
// the maginit routines.
/*********************************************
#ifdef PERTURBATION_SIZE
#undef PERTURBATION_SIZE
#endif
#define PERTURBATION_SIZE 0.
**********************************************/


// NOTE: To add new magnetization initialization functions, add the name
// to the following list, and fill in the MagInit::Ident structure below.
// Then use the existing functions as a guide to writing your own function.

MI_Function MI_InUpOut,MI_InLeftOut,MI_InLeftOutRot,MI_InWideUpOut;
MI_Function MI_InOut,MI_InOutRot,MI_InUpOutRot,MI_CRot;
MI_Function MI_In,MI_Up;
MI_Function MI_UpDownHeadToHead;
MI_Function MI_UpOutDownHeadToHead,MI_UpRightDownHeadToHead;
MI_Function MI_InOuts,MI_UpDowns;
MI_Function MI_UpDownsRot,MI_UpDownShift;
MI_Function MI_Cant,MI_Random,MI_Bloch;
MI_Function MI_4domain,MI_1domain,MI_7domain;
MI_Function MI_Vortex,MI_Exvort;
MI_Function MI_Sphere,MI_Source;
MI_Function MI_RightLeft,MI_RightUpLeft;
MI_Function MI_vioFile;

MI_FuncInfo MagInit::Ident[]={
  { MI_Random,	    "Random",	    0}, // Default; MUST require no parameters
  { MI_InUpOut,	    "InUpOut",	    0},
  { MI_InLeftOut,   "InLeftOut",    0},
  { MI_InLeftOutRot,"InLeftOutRot", 1},
  { MI_InWideUpOut, "InWideUpOut",  0},
  { MI_InOut,	    "InOut",	    0},
  { MI_InOutRot,    "InOutRot",	    1},
  { MI_InUpOutRot,  "InUpOutRot",   1},
  { MI_CRot,        "CRot",         1},
  { MI_In,	    "In",	    0},
  { MI_Up,	    "Up",	    0},
  { MI_UpDownHeadToHead,    "UpDownHeadToHead",    0},
  { MI_UpOutDownHeadToHead, "UpOutDownHeadToHead", 1},
  { MI_UpRightDownHeadToHead, "UpRightDownHeadToHead", 1},
  { MI_InOuts,	    "InOuts",	    1},
  { MI_UpDowns,	    "UpDowns",	    1},
  { MI_UpDownsRot,  "UpDownsRot",   2},
  { MI_UpDownShift, "UpDownShift",  1},
  { MI_Bloch,	    "Bloch",	    1},
  { MI_Cant,	    "Cant",	    2},
  { MI_Vortex,	    "Vortex",	    0},
  { MI_Exvort,	    "Exvort",	    0},
  { MI_Sphere,	    "Sphere",	    0},
  { MI_Source,	    "Source",	    0},
  { MI_1domain,	    "1domain",	    0},
  { MI_4domain,	    "4domain",	    0},
  { MI_7domain,	    "7domain",	    0},
  { MI_RightLeft,   "RightLeft",    0},
  { MI_RightUpLeft, "RightUpLeft",  0},
  { MI_vioFile,     "vioFile",      1}
};

int MagInit::IdentCount=0;
static MagInit maginit_dummy; // Dummy variable to run initializer

char* MagInit::NameCheck(char *index)
{ // Returns pointer to routine name, or NULL if no match found
  int i;
  if(index==NULL || index[0]=='\0') return Ident[0].name;
  for(i=0;i<IdentCount;i++)
    if(Omf_StrCaseCmp(Ident[i].name,index)==0) break;
  if(i>=IdentCount) return (char*)NULL;
  return Ident[i].name;
}

MI_Function* MagInit::LookUp(char *index)
{
  int i;
  if(index==NULL || index[0]=='\0') return Ident[0].func;
  for(i=0;i<IdentCount;i++)
    if(Omf_StrCaseCmp(Ident[i].name,index)==0) break;
  if(i>=IdentCount)
    PlainError(1,"Error in MagInit::Lookup: Magnetization index string not matched");
  return Ident[i].func;
}

char* MagInit::LookUp(MI_Function* func)
{
  int i;
  for(i=0;i<IdentCount;i++) if(Ident[i].func==func) break;
  if(i>=IdentCount) return (char*)NULL;
  return Ident[i].name;
}

int MagInit::GetParamCount(char *index)
{
  int i;
  if(index==NULL || index[0]=='\0') return Ident[0].paramcount;
  for(i=0;i<IdentCount;i++)
    if(Omf_StrCaseCmp(Ident[i].name,index)==0) break;
  if(i>=IdentCount)
    PlainError(1,"Error in MagInit::GetParamCount: "
	       "Magnetization index string ->%s<- not matched",index);
  return Ident[i].paramcount;
}

void MagInit::List(FILE* output)
{
  fprintf(output,"%d magnetization initialization routines available---\n",
	  IdentCount);
  int i;
  for(i=0;i<IdentCount;i++) {
    if(Ident[i].name[0]!='\0')
      fprintf(output,"%11s with %d arguments",
	      Ident[i].name,Ident[i].paramcount);
    if(i%3 == 2) fprintf(output,"\n"); else fprintf(output," ");
  }
  if(i%3 != 3) fprintf(output,"\n");
}


MagInit::MagInit(void)
{
  IdentCount=sizeof(Ident)/sizeof(MI_FuncInfo);
}

int MI_Random(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j].Random(1.0);
  }
  return 0;
}

int MI_InOut(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    if(i<Nx/2) m[i][j][1]=1.0;
    else       m[i][j][1]=-1.0;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_InOutRot(int Nx,int Nz,MagElt** m,char** params)
{ // Same as InOut, but rotated params[0] degrees ccw in xz-plane
  const int ParamCount=1; // This routine requires 1 argument: Phi
  double phi;  // Parameters
  int i,j;

  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_InOutRot: This magnetization initialization function"
	       " requires %d argument: Phi (in degrees); None passed", ParamCount);

  // Read parameter
  phi=atof(params[0]);
  printf("Initializing In-Out magnetization configuration at angle"
	 " phi=%5.3f (degrees)\n",phi);
  phi*=PI/180.; // Convert to radians

  double cosphi,sinphi,offset;
  cosphi=cos(phi); sinphi=sin(phi);
  offset=double(Nx-1)*cosphi/2.+double(Nz-1)*sinphi/2.;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    if(double(i)*cosphi+double(j)*sinphi<offset) m[i][j][1]=  1.0;
    else                                         m[i][j][1]= -1.0;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_InUpOutRot(int Nx,int Nz,MagElt** m,char** params)
{ // Same as InUpOut, but rotated params[0] degrees ccw in xz-plane
  const int ParamCount=1; // This routine requires 1 argument: Phi
  double phi;  // Parameters
  int i,j;

  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_InOutRot: This magnetization initialization function"
	       " requires %d argument: Phi (in degrees); None passed", ParamCount);

  // Read parameter
  phi=atof(params[0]);
  printf("Initializing In-Out magnetization configuration at angle"
	 " phi=%5.3f (degrees)\n",phi);
  phi*=PI/180.; // Convert to radians

  double cosphi,sinphi,offset,center;
  cosphi=cos(phi); sinphi=sin(phi);
  center=double(Nx-1)*cosphi/2.+double(Nz-1)*sinphi/2.;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    offset=double(i)*cosphi+double(j)*sinphi;
    if(offset<center-0.5)      m[i][j][1]=  1.0;
    else if(offset>center+0.5) m[i][j][1]= -1.0;
    else { // Center region
      m[i][j][0]= -sinphi;  m[i][j][1]=0.0;  m[i][j][2]=cosphi;
    }

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_CRot(int Nx,int Nz,MagElt** m,char** params)
{ // Same as InUpOutRot, but top and bottom spins rotated to form a "C",
  const int ParamCount=1; // This routine requires 1 argument: Phi
  double phi;  // Parameters
  int i,j;

  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_CRot: This magnetization initialization function"
	       " requires %d argument: Phi (in degrees); None passed", ParamCount);

  if(Nz<2 || Nx<2)
    PlainError(1,"Error in MI_CRot: This magnetization initialization function "
	       "requires both Nx (=%d) and Nz (=%d) >1.",Nx,Nz);

  // Read parameter
  phi=atof(params[0]);
  printf("Initializing In-Out magnetization configuration at angle"
	 " phi=%5.3f (degrees)\n",phi);
  phi*=PI/180.; // Convert to radians

  double cosphi,sinphi,offset,center,height,centerheight,maxheight,voff;
  cosphi=cos(phi); sinphi=sin(phi);
  center=double(Nx-1)*cosphi/2.+double(Nz-1)*sinphi/2.;
  centerheight=double(Nx-1)*sinphi/2.-double(Nz-1)*cosphi/2.;
  if(fabs((Nz-1)*sinphi)<fabs((Nx-1)*cosphi))
    maxheight=fabs(double(Nz-1)/(2*cosphi));
  else
    maxheight=fabs(double(Nx-1)/(2*sinphi));

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    offset=double(i)*cosphi+double(j)*sinphi;
    height=double(i)*sinphi-double(j)*cosphi-centerheight;
    voff=maxheight-fabs(height);
    if(offset<center-0.5) m[i][j][1]=1.0;
    else if((offset>center+1.5) ||
	    (offset>center+0.5 && voff>1)) m[i][j][1]= -1.0;
    else { // Center region
      m[i][j][1]=0.0;
      if(voff>1) { // middle
	m[i][j][0]= -sinphi;  m[i][j][2]=  cosphi;
      }
      else if(height>0) { // top
	m[i][j][0]= -cosphi;  m[i][j][2]= -sinphi;
      }
      else { // bottom
	m[i][j][0]=  cosphi;  m[i][j][2]=  sinphi;
      }
    }

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_InUpOut(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    if(i<Nx/2)	           m[i][j][1]=  1.0;
    else if(i>Nx/2)        m[i][j][1]= -1.0;
    else { m[i][j][1]=0.0; m[i][j][2]=  1.0; }
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_InLeftOut(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    if(i<Nx/2)	           m[i][j][1]=  1.0;
    else if(i>Nx/2)        m[i][j][1]= -1.0;
    else { m[i][j][1]=0.0; m[i][j][0]= -1.0; }
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_InLeftOutRot(int Nx,int Nz,MagElt** m,char** params)
{ // Same as InLeftOut, but rotated params[0] degrees ccw in xz-plane
  const int ParamCount=1; // This routine requires 1 argument: Phi
  double phi;  // Parameters
  int i,j;

  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_InLeftOutRot: This magnetization initialization function"
	       " requires %d argument: Phi (in degrees); None passed",
	       ParamCount);

  // Read parameter
  phi=atof(params[0]);
  printf("Initializing In-Out magnetization configuration at angle"
	 " phi=%5.3f (degrees)\n",phi);
  phi*=PI/180.; // Convert to radians

  double cosphi,sinphi,offset,center;
  cosphi=cos(phi); sinphi=sin(phi);
  center=double(Nx-1)*cosphi/2.+double(Nz-1)*sinphi/2.;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    offset=double(i)*cosphi+double(j)*sinphi;
    if(offset<center-0.5)      m[i][j][1]=  1.0;
    else if(offset>center+0.5) m[i][j][1]= -1.0;
    else { // Center region
      m[i][j][0]= -cosphi;  m[i][j][1]=0.0;  m[i][j][2]= -sinphi;
    }

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_InWideUpOut(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    if(i<(2*Nx)/5)	   m[i][j][1]=1.0;
    else if(i>Nx-(2*Nx)/5) m[i][j][1]=-1.0;
    else {
      double theta;
      theta= PI/2-PI*double(i-(2*Nx)/5)/double(Nx-2*((2*Nx)/5));
      m[i][j][1]=sin(theta); m[i][j][2]=cos(theta);
    }
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_In(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][2]=0.0;
    m[i][j][1]=1.0;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_Up(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][1]=0.0;
    m[i][j][2]=1.0;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_UpDownHeadToHead(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][1]=0.0;
    if(j<Nz/2) m[i][j][2]=  1.0;
    else       m[i][j][2]= -1.0;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_UpOutDownHeadToHead(int Nx,int Nz,MagElt** m,char** params)
{
  const int ParamCount=1; // This routine requires 1 argument: Out proportion
  double out_proportion;
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_UpOutDownHeadToHead: "
	       "This magnetization initialization function"
	       " requires %d argument: Out proportion; None passed",
	       ParamCount);
  // Read parameter
  out_proportion=atof(params[0]);
  if(out_proportion<0 || out_proportion>1)
    PlainError(1,"Error in MI_UpOutDownHeadToHead: "
	       "This magnetization initialization function requires %d argument:"
	       " Out proportion; Illegal value (%d) passed (should be in [0,1])",
	       ParamCount,out_proportion);
  int outstart,outstop;
  outstart=ROUND(Nz*(1.-out_proportion)/2.);
  outstop=Nz-outstart;

  int i,j;
  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][1]=m[i][j][2]=0.0;
    if(j<outstart)      m[i][j][2]=  1.0;
    else if(j<outstop)  m[i][j][1]= -1.0;
    else                m[i][j][2]= -1.0;
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_UpRightDownHeadToHead(int Nx,int Nz,MagElt** m,char** params)
{
  const int ParamCount=1; // This routine requires 1 argument: Right proportion
  double right_proportion;
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_UpRightDownHeadToHead: "
	       "This magnetization initialization function"
	       " requires %d argument: Right proportion; None passed",
	       ParamCount);
  // Read parameter
  right_proportion=atof(params[0]);
  if(right_proportion<0 || right_proportion>1)
    PlainError(1,"Error in MI_UpRightDownHeadToHead: "
	       "This magnetization initialization function requires %d argument:"
	       " Right proportion; Illegal value (%d) passed (should be in [0,1])",
	       ParamCount,right_proportion);
  int rightstart,rightstop;
  rightstart=ROUND(Nz*(1.-right_proportion)/2.);
  rightstop=Nz-rightstart;

  int i,j;
  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][1]=m[i][j][2]=0.0;
    if(j<rightstart)      m[i][j][2]=  1.0;
    else if(j<rightstop)  m[i][j][0]=  1.0;
    else                  m[i][j][2]= -1.0;
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_UpDowns(int Nx,int Nz,MagElt** m,char** params)
{
  const int ParamCount=1; // This routine requires 1 argument: Domain width
  int domain_width;	  // Parameters
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_UpDowns: This magnetization initialization function"
	       " requires %d argument: Domain width (in cells); None passed",
	       ParamCount);

  // Read parameter
  domain_width=atoi(params[0]);
  if(domain_width<1)
    PlainError(1,"Error in MI_UpDowns: "
	       "This magnetization initialization function requires %d argument:"
	       " Domain width (in cells); Illegal value (%d) passed",
	       ParamCount,domain_width);
  printf("Initializing with Up and Down domains %d cells wide\n",domain_width);

  int i,j;
  double mag;

  mag=-1.0;
  for(i=0;i<Nx;i++) {
    if(i%domain_width==0) mag*=-1;
    for(j=0;j<Nz;j++) {
      m[i][j][0]=m[i][j][1]=0.0;
      m[i][j][2]=mag;
      // Adjust with random perturbations
      m[i][j].Perturb(PERTURBATION_SIZE);
      m[i][j].Scale(1.);
    }
  }
  return 0;
}

int MI_UpDownsRot(int Nx,int Nz,MagElt** m,char** params)
{ // Similar to MI_UpDowns, but rotated.
  const int ParamCount=2; // This routine requires 2 arguments:
  int domain_width;	  //   domain width and orientation angle
  double orient;
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_UpDowns: This magnetization initialization function"
	       " requires %d arguments:\n"
	       " Domain width (in cells) and orientation angle (in degrees);"
	       " None passed", ParamCount);

  // Read parameters
  domain_width=atoi(params[0]);
  if(domain_width<1)
    PlainError(1,"Error in MI_UpDownsRot: "
	       "Illegal domain width value (%d cells) requested.\n",
	       domain_width);
  orient=atof(params[1]);
  printf("Initializing with Up and Down domains %d cells wide,\n"
	 " oriented %f degrees from vertical\n",domain_width,orient);
  orient*=PI/180.;

  int i,k,ioffset;
  double xc,zc,vx,vz,magx,magz,offx,offset,wall;

  xc=(Nx-1)/2.;  zc=(Nz-1)/2.;
  vx=cos(orient); vz=sin(orient);
  magx= -vz; magz=vx;

  for(i=0;i<Nx;i++) {
    offx=(double(i)-xc)*vx;
    for(k=0;k<Nz;k++) {
      m[i][k][1]=0.0; // All moments in x-z plane
      offset=(offx+(double(k)-zc)*vz)/domain_width+0.5;
      ioffset=int(floor(offset));
      if(ioffset%2==0) { m[i][k][0]=  magx; m[i][k][2]=  magz; }
      else             { m[i][k][0]= -magx; m[i][k][2]= -magz; }
      wall=(offset-ioffset); if (wall>0.5) wall--;
      wall*=domain_width;
      if(-0.9<wall && wall<=0.35) { // At wall center;
	m[i][k][0]=magz; m[i][k][2]=-magx;      // Rotate horizontally
      }
      // Adjust with random perturbations
      m[i][k].Perturb(PERTURBATION_SIZE);
      m[i][k].Scale(1.);
    }
  }
  return 0;
}

int MI_UpDownShift(int Nx,int Nz,MagElt** m,char** params)
{
  const int ParamCount=1; // This routine requires 1 argument: Wall shift,
  double wall_shift;	  // in [-.5,.5], with 0 => wall center at Nx/2.
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_UpDownShift: "
	       "This magnetization initialization function"
	       " requires %d argument: Wall shift (proportion;"
	       " 0==centered); None passed",ParamCount);

  // Read parameter
  wall_shift=atof(params[0]);
  if(wall_shift<-0.5 || 0.5<wall_shift)
    PlainError(1,"Error in MI_UpDownShift: "
	       "This magnetization initialization function"
	       " requires %d argument: Wall shift ([-.5,.5], "
	       " 0==centered); Passed value: %f",ParamCount,wall_shift);
  printf("Initializing with Up and Down domain shifted by %f\n",
	 wall_shift);

  int i,j;
  double mag,mx,my,mz;
  int midbreak=int(rint(double(Nx)*(0.5+wall_shift)));
  if(midbreak<0) midbreak=0;
  else if(midbreak>Nx) midbreak=Nx;

  mag=1.0;
  mx=0.0; my=0.0; mz=mag;
  for(i=0;i<Nx;i++) {
    if(i==midbreak) {
      mx=0.0; my=mag; mz=0.0;
    }
    else if(i==midbreak+1) {
      mag*=-1;
      mx=0.0; my=0.0; mz=mag;
    }
    for(j=0;j<Nz;j++) {
      m[i][j][0]=mx;
      m[i][j][1]=my;
      m[i][j][2]=mz;
      // Adjust with random perturbations
      m[i][j].Perturb(PERTURBATION_SIZE);
      m[i][j].Scale(1.);
    }
  }
  return 0;
}

int MI_InOuts(int Nx,int Nz,MagElt** m,char** params)
{
  const int ParamCount=1; // This routine requires 1 argument: Domain width
  int domain_width;	  // Parameters
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_InOuts: This magnetization initialization"
	       " function requires %d argument: Domain width (in cells);"
	       " None passed", ParamCount);
  // Read parameter
  domain_width=atoi(params[0]);
  if(domain_width==0)
    PlainError(1,"Error in MI_InOuts: "
	       "This magnetization initialization function requires"
	       " %d argument: Domain width (in cells); Illegal value"
	       " (%d) passed",ParamCount,domain_width);
  printf("Initializing with In and Out domains %d cells wide,",
	 abs(domain_width));
  if(domain_width>0) printf(" iterated left to right\n");
  else               printf(" iterated top to bottom\n");

  int i,j;
  double mag;

  mag=-1.0;
  if(domain_width>0) { // Iterate domains from left to right
    for(i=0;i<Nx;i++) {
      if(i%domain_width==0) mag*=-1;
      for(j=0;j<Nz;j++) {
	m[i][j][0]=m[i][j][2]=0.0;
	m[i][j][1]=mag;
	// Adjust with random perturbations
	m[i][j].Perturb(PERTURBATION_SIZE);
	m[i][j].Scale(1.);
      }
    }
  }
  else { // Iterate domains from top to bottom
    domain_width*=-1;
    for(j=0;j<Nz;j++) {
      if(j%domain_width==0) mag*=-1;
      for(i=0;i<Nx;i++) {
	m[i][j][0]=m[i][j][2]=0.0;
	m[i][j][1]=mag;
	// Adjust with random perturbations
	m[i][j].Perturb(PERTURBATION_SIZE);
	m[i][j].Scale(1.);
      }
    }
  }
  return 0;
}

int MI_Bloch(int Nx,int Nz,MagElt** m,char** params)
{ // InUpOut rotated about the x axis
  const int ParamCount=1; // This routine requires 1 argument: Phi
  double phi;  // Parameters
  int i,j;

  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_Bloch: "
	       "This magnetization initialization function requires"
	       " %d argument: Phi (in degrees); None passed", ParamCount);

  // Read parameter
  phi=atof(params[0]);
  printf("Initializing Bloch wall magnetization at angle"
	 " phi=%5.3f (degrees)\n",phi);
  phi*=PI/180.; // Convert to radians

  double x,y,z;
  x=  0.0;
  y=  cos(phi);
  z=  sin(phi);

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=x;
    if(i<Nx/2)	     { m[i][j][1]=  y;  m[i][j][2]= z; }
    else if(i==Nx/2) { m[i][j][1]= -z;  m[i][j][2]= y; }
    else	     { m[i][j][1]= -y;  m[i][j][2]=-z; }

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_Cant(int Nx,int Nz,MagElt** m,char** params)
{
  const int ParamCount=2; // This routine requires 2 arguments: Theta and Phi
  double theta,phi;  // Parameters
  int i,j;

  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_Cant: "
	       "This magnetization initialization function requires %d arguments:"
	       " Theta and Phi (in degrees); None passed", ParamCount);
  for(i=0;i<ParamCount;i++) if(params[i][0]=='\0')
    PlainError(1,"Error in MI_Cant: "
	       "This magnetization initialization function requires %d arguments:"
	       " Theta and Phi (in degrees); Only %d passed", ParamCount,i);

  // Read parameters
  theta=atof(params[0]); phi=atof(params[1]);
  printf("Initializing magnetization in direction"
	 " theta=%5.3f, phi=%5.3f (degrees)\n",
	 theta,phi);
  theta*=PI/180.; phi*=PI/180.; // Convert to radians
  double x,y,z;
  x=cos(theta)*sin(phi);
  y=sin(theta)*sin(phi);
  z=cos(phi);

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=x; m[i][j][1]=y; m[i][j][2]=z;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_1domain(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][0]=m[i][j][1]=0.0; m[i][j][2]=1.0;
    if( j<Nx-i )
      { m[i][j][0] = 0.7; m[i][j][2] = 0.7;}
    if( j > Nz-i )
      { m[i][j][0] = 0.7; m[i][j][2] = 0.7;}
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_4domain(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  if ( Nz > Nx ) {
    for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
      m[i][j][0]=m[i][j][1]=  0.0;
      if(i<Nx/2) m[i][j][2]=  1.0;
      else	 m[i][j][2]= -1.0;
      if( j<i && j<Nx-1-i )
	{ m[i][j][0] = -1.0; m[i][j][2] =  0.0;}
      if( j > Nz-1-i && j > Nz-Nx+i)
	{ m[i][j][0] =  1.0; m[i][j][2] =  0.0;}
      if( j == i && i < Nx/2 )
	{ m[i][j][0] = -0.7; m[i][j][2] =  0.7;}
      if( j == Nx-1-i && i >= Nx/2 )
	{ m[i][j][0] = -0.7; m[i][j][2] = -0.7;}
      if( j == Nz-1-i && i < Nx/2 )
	{ m[i][j][0] =  0.7; m[i][j][2] =  0.7;}
      if( j == Nz-Nx+i && i >= Nx/2 )
	{ m[i][j][0] =  0.7; m[i][j][2] = -0.7;}
      // Adjust with random perturbations
      m[i][j].Perturb(PERTURBATION_SIZE);
      m[i][j].Scale(1);
    }
  }
  else {
    for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
      m[i][j][0]=m[i][j][1]=  0.0;
      if(j<Nz/2) m[i][j][0]=  1.0;
      else	 m[i][j][0]= -1.0;
      if( i<j && i<Nz-1-j )
	{ m[i][j][2] = -1.0; m[i][j][0] =  0.0;}
      if( i > Nx-1-j && i > Nx-Nz+j)
	{ m[i][j][2] =  1.0; m[i][j][0] =  0.0;}
      if( i == j && j < Nz/2 )
	{ m[i][j][2] = -0.7; m[i][j][0] =  0.7;}
      if( i == Nz-1-j && j >= Nz/2 )
	{ m[i][j][2] = -0.7; m[i][j][0] = -0.7;}
      if( i == Nx-1-j && j < Nz/2 )
	{ m[i][j][2] =  0.7; m[i][j][0] =  0.7;}
      if( i == Nx-Nz+j && j >= Nz/2 )
	{ m[i][j][2] =  0.7; m[i][j][0] = -0.7;}
      // Adjust wjth random perturbations
      m[i][j].Perturb(PERTURBATION_SIZE);
      m[i][j].Scale(1);
    }
  }
  return 0;
}

int MI_7domain(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;
  if (Nz > Nx) {
    for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
      if( j<Nz/4 || j > 3*Nz/4)			// center domains
	{ m[i][j][0] = -1.0; m[i][j][1]=m[i][j][2]=0.0; }
      else
	{ m[i][j][0] =  1.0; m[i][j][1]=m[i][j][2]=0.0; }
      if( j>i && j<Nx-i )			// 4 edge domains
	{ m[i][j][2] =  1.0; m[i][j][0]=m[i][j][1]=0.0; }
      if( j<Nz-Nx+i && j>Nz-i )
	{ m[i][j][2] =  1.0; m[i][j][0]=m[i][j][1]=0.0; }
      if( j<i && j>Nx-i )
	{ m[i][j][2] = -1.0; m[i][j][0]=m[i][j][1]=0.0; }
      if( j>Nz-Nx+i && j<Nz-i )
	{ m[i][j][2] = -1.0; m[i][j][0]=m[i][j][1]=0.0; }
      // Adjust with random perturbations
      m[i][j].Perturb(PERTURBATION_SIZE);
      m[i][j].Scale(1);
    }
  }
  else {
    for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
      if( i<Nx/4 || i > 3*Nx/4)			// center domains
	{ m[i][j][0] = -1.0; m[i][j][1]=m[i][j][2]=0.0; }
      else
	{ m[i][j][0] =  1.0; m[i][j][1]=m[i][j][2]=0.0; }
      if( i>j && i<Nz-j )				// 4 edge domains
	{ m[i][j][2] =  1.0; m[i][j][0]=m[i][j][1]=0.0; }
      if( i<Nx-Nz+j && i>Nx-j )
	{ m[i][j][2] =  1.0; m[i][j][0]=m[i][j][1]=0.0; }
      if( i<j && i>Nz-j )
	{ m[i][j][2] = -1.0; m[i][j][0]=m[i][j][1]=0.0; }
      if( i>Nx-Nz+j && i<Nx-j )
	{ m[i][j][2] = -1.0; m[i][j][0]=m[i][j][1]=0.0; }

      // Adjust with random perturbations
      m[i][j].Perturb(PERTURBATION_SIZE);
      m[i][j].Scale(1);
    }
  }
  return 0;
}

int MI_Vortex(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  if(Nx<1 || Nz<1 || m==(MagElt**)NULL)
    PlainError(1,"Error in MI_Vortex: %s",ErrBadParam);
  int i,j;
  float x,z;
  float midx, midz;

  midx = (Nx-1)/2.0; midz = (Nz-1)/2.0;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    if ( midx != (float) i || midz != (float) j ) {
      x = i-midx; z = j-midz;
      m[i][j][0] = -z/hypot(x,z);
      m[i][j][2] =  x/hypot(x,z);
      m[i][j][1] =  0.0;
    }
    else
      { m[i][j][0]=m[i][j][2]=0.0; m[i][j][1] = 1.0;}
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_Exvort(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;
  float x,z;
  float midx, midz;

  midx = (Nx-1)/2.0; midz = (Nz-1)/2.0;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    if ( midx != (float) i || midz != (float) j ) {
      x = i-midx; z = j-midz;
      m[i][j][0] =  z/hypot(x,z);
      m[i][j][2] =  x/hypot(x,z);
      m[i][j][1] =  0.0;
    }
    else
      { m[i][j][0]=m[i][j][2]=0.0; m[i][j][1] = 1.0;}
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_Sphere(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  if(Nx<1 || Nz<1 || m==(MagElt**)NULL)
    PlainError(1,"Error in MI_Sphere: %s",ErrBadParam);
  int i,j;
  float x,y,z,proj_len;
  float midx, midz, radius, radius_sq;

  midx = (Nx-1)/2.0; midz = (Nz-1)/2.0;
  radius = MAX(MIN(midx,midz),1.0);
  radius_sq=SQ(radius);

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    x = i-midx; z = j-midz;
    proj_len=hypot(x,z);
    if(proj_len >= radius) { // Outside sphere
      m[i][j][0] =  x/proj_len;
      m[i][j][2] =  z/proj_len;
      m[i][j][1] =  0.0;
    }
    else {
      y=sqrt(radius_sq-SQ(proj_len));
      m[i][j][0] =  x/radius;
      m[i][j][2] =  z/radius;
      m[i][j][1] =  y/radius;
    }
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_Source(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  if(Nx<1 || Nz<1 || m==(MagElt**)NULL)
    PlainError(1,"Error in MI_Source: %s",ErrBadParam);
  int i,j;
  float x,z;
  float midx, midz;

  midx = (Nx-1)/2.0; midz = (Nz-1)/2.0;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    if ( midx != (float) i || midz != (float) j ) {
      x = i-midx; z = j-midz;
      m[i][j][0] =  x/hypot(x,z);
      m[i][j][2] =  z/hypot(x,z);
      m[i][j][1] =  0.0;
    }
    else
      { m[i][j][0]=m[i][j][2]=0.0; m[i][j][1] = 1.0;}
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_RightLeft(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;

  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][1]=m[i][j][2]= 0.0;
    if(i<Nx/2) m[i][j][0]= 1.0;
    else       m[i][j][0]=-1.0;

    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE); 
    m[i][j].Scale(1);	  
  }
  return 0;
}

int MI_RightUpLeft(int Nx,int Nz,MagElt** m,char** /*params*/)
{
  int i,j;
  for(i=0;i<Nx;i++) for(j=0;j<Nz;j++) {
    m[i][j][1]=m[i][j][2]=0.0;
    if(i<(Nx/2-j))      m[i][j][0]=  1.0;
    else if(i>(Nx/2-j)) m[i][j][0]= -1.0;
    else {
      if (j>Nz/2) m[i][j][2]=  -1.0; 
      else        m[i][j][2]=   1.0;
    }
    // Adjust with random perturbations
    m[i][j].Perturb(PERTURBATION_SIZE);
    m[i][j].Scale(1);
  }
  return 0;
}

int MI_vioFile(int Nx,int Nz,MagElt** m,char** params)
{ // Initializes magnetization from *.vio file.  If dimensions of
  // file are not the same as the grid (i.e., Nx & Nz), then stretch
  // and/or shrink the input file and interpolate using 0th order fit.

  const int ParamCount=1; // This routine requires 1 argument: Phi
  // Do a parameter list length check
  if(params==NULL || params[0][0]=='\0'
     || params[0][0]=='?' || params[0][1]=='?')
    PlainError(1,"Error in MI_vioFile: "
	       "This magnetization initialization function requires %d argument:"
	       " filename.vio; None passed", ParamCount);

  int xfiledim,yfiledim,zfiledim,vd;
  FileVector fv;
  // Read file header and get sizing information
  if(fv.ReadHeader(params[0],xfiledim,yfiledim,zfiledim,vd)!=0 || yfiledim!=1)
    PlainError(1,"Error in MI_vioFile: "
	       "Bad or nonexistent *.vio input file: \"%s\".",params[0]);

  // Check that spin dimensions agree with those from input file
  if(vd!=VECDIM)
    PlainError(1,"Error in MI_vioFile: "
	       "Magnetization dimension (%d) incompatible with that of"
	       " input file %s (%d)",VECDIM,params[0],vd);

  // See if grid dimensions agree with those from input file
  double xoff,xscale,zoff,zscale; // Linear parameters to convert from
                                  // mesh coordinates to file coordinates.
  if(xfiledim!=Nx || zfiledim!=Nz) {
    PlainWarning("Error in MI_vioFile: \n"
		 "  Grid dimensions (%dx%d) are different with those"
		 " from input file %s (%dx%d)\n  Will stretch and/or"
		 " shrink input as necessary.  Results may be"
		 " sub-optimal!\n",Nx,Nz,params[0],xfiledim,zfiledim);
    xoff=double(xfiledim-Nx)/double(2*Nx);
    xscale=double(xfiledim)/double(Nx);
    zoff=double(zfiledim-Nz)/double(2*Nz);
    zscale=double(zfiledim)/double(Nz);
  }
  else {
    xoff=zoff=0.0;
    xscale=zscale=1.0;
  }

  // Allocate workspace memory
  double *arr;
  if(!(arr=new double[zfiledim*vd])) 
    PlainError(1,"Error in MI_vioFile: %s",ErrNoMem);

  // Fill array
  int itemp,j,imesh,kmesh,ifile,kfile;
  for(ifile=imesh=0;ifile<xfiledim && imesh<Nx;ifile++) {
    fv.ReadVecs(zfiledim,arr); // Read in 1 row from file
    while(imesh<Nx) { // Fill all appropriate mesh rows from this file row
      itemp=ROUND(xoff+xscale*imesh);
      if(itemp<0)         itemp=0;          // Safety
      if(itemp>=xfiledim) itemp=xfiledim-1; // Safety
      if(itemp!=ifile) break;
      for(kmesh=0;kmesh<Nz;kmesh++) {
	kfile=ROUND(zoff+zscale*kmesh);
	if(kfile<0)         kfile=0;          // Safety
	if(kfile>=zfiledim) kfile=zfiledim-1; // Safety
	for(j=0;j<VECDIM;j++) m[imesh][kmesh][j]=arr[kfile*VECDIM+j];
      }
      imesh++;
    }
  }

  // Clean up and exit
  delete arr;
  fv.Close();
  return 0;
}
