# FILE: channellog.tcl
#
# Each instance of the em_class net_channellog writes a record of
# text messages to a writable channel.
#
# Need way to register dependence on continued existence of open channel.
#
# Currently, the 'puts' which writes the messages out to the channel
# will block when the output buffer gets full while the data is 
# transferred to the underlying device.  If that causes unwelcome
# delays, we'll need to use a non-blocking output channel, along
# with some internal buffering here, and appropriate use of the
# fileevent writable callback.  For now, keep it simple.
#
# A good candidate for inheritance.
#

# Only define em_channellog once (bootstrap protection)
if {[string match {} [info procs em_channellog]]} {

em_class em_channellog {
    common serial 0
    public variable channel
    private variable imadechannel = 0

    public variable title = "Channel Log"
    public variable dateformat = "\[%Y %b %d, %H:%M:%S %Z\] "
    private variable lineslogged

    private variable bindtags

    constructor {args} {
        set lineslogged 0
        set bindtags [list $this $class]
        eval $this configure $args
        if {![info exists channel]} {
            # A default channel for logs
            global tcl_platform DEBUGLVL
            if {[info exists DEBUGLVL] && ($DEBUGLVL == 1) \
                    && [string match unix $tcl_platform(platform)]} {
                # For debugging on Unix development platforms, it's convenient
                # to spill messages to stderr where they can be seen with
                # little todo.
                set channel stderr
            } else {
                # In other situations, its better to direct log information
                # to a file, where it won't be so intrusive, and where it
                # will persist for later review even after the application
                # has terminated.
                #	Should prefix log filename with base derived
                # from name of running application.  How to get that?
                set channel [open [pid]-$serial.log a]
                set imadechannel 1
                incr serial
            }
        }
        # Make sure each line gets flushed as we go
        fconfigure $channel -buffering line
        $this Append "Starting $title ..."
    }
 
    # Append a line to this log
    method Append { args } {
        if {[llength $args] == 1} {
            set msg [lindex $args 0]
            set nonewline 0
        } elseif {([llength $args] == 2) \
                && [string match -nonewline [lindex $args 0]]} {
            set msg [lindex $args 1]
            set nonewline 1
        } else {
            error "usage: <em_channellog instance> Append ?-nonewline? <msg>"
        }
        # Workaround of bug in clock command in Tcl 7.5 - 7.6p2
        if {[regexp % $dateformat]} {
            set msg "[clock format [clock seconds] -format $dateformat]$msg"
        } else {
            set msg $dateformat$msg
        }
        incr lineslogged [llength [split $msg "\n"]]
        if {$nonewline} {
            puts -nonewline $channel "$msg"
        } else {
            puts $channel "$msg"
        }
    }

    destructor {
        em_eventHandler Generate delete $bindtags
        flush $channel
        if {$imadechannel} {
            close $channel
        }
    }

}

}
