# FILE: graphbox.tcl
#
# Entry graph canvas sub-widget
#
# debugging enable/disable/level

em_class omf_graphwin {
    const public variable winpath
    public variable curve_width = 2  ;# Width of curve plot lines
    const public variable outer_frame_options = \
	    "-bd 0 -relief ridge -background #0000cd"
    # Note: #0000cd is MediumBlue
    const public variable default_height = 320
    const public variable default_width = 320
    const public variable canvas_frame_options = \
	    "-bd 0 -relief ridge -background #0000cd"
    const public variable canvas_options = \
	    "-bd 0 -relief flat -background white"
    const public variable graphsize = 0.9  ;# % of canvas to draw graph in.
    const public variable axescolor = black
    const public variable axeswidth = 2
    const public variable keywidth = 2

    const private variable graph  ;# Graph canvas

    private variable xlabel = ""
    private variable ylabel = ""

    public variable autodraw = 1
    public variable autoscale = 1
    public variable showkey = 1

    private variable curvecount = 0
    private variable xorg    = 0
    private variable yorg    = 0
    private variable xscale  = 1
    private variable yscale  = 1
    private variable lmargin = 0    ;# Margins: left, top, right, bottom
    private variable tmargin = 0
    private variable rmargin = 0
    private variable bmargin = 0
    private variable xoff    = 0    ;# Used in conversion from data coords
    private variable yoff    = 0    ;#  to graph coords
    private variable xgraphmin = 0
    private variable xgraphmax = 1
    private variable ygraphmin = 0
    private variable ygraphmax = 1

    private variable but1x = 0       ;# Mouse coordinates when
    private variable but1y = 0       ;# Button-1 is pressed

    private variable color           ;# Color array

    private variable id              ;# Name -> id mapping array
    private variable curve           ;# Curve data array
    ## Each curve has an entry of the following form:
    ##       curve($curveid,name)    = id -> Name mapping
    ##       curve($curveid,color)   = Drawing color
    ##       curve($curveid,pts)     = Data point string of "x y" pairs
    ##       curve($curveid,xmin)    = Minimum value of x among data points
    ##       curve($curveid,xmax)    = Maximum value of x among data points
    ##       curve($curveid,ymin)    = Minimum value of y among data points
    ##       curve($curveid,ymax)    = Maximum value of y among data points
    ##       curve($curveid,ptsdisp) = Display point string of "x y" pairs
    ##       curve($curveid,lastdisp) = Last element pair in ptsdisp list
    ##       curve($curveid,scalevalid) = True if ptsdisp is properly
    ##                                 scaled for current graph.
    ##       curve($curveid,hidden)  = Controls whether or not curve is drawn
    ##       curve($curveid,keyid)   = Canvas id associated with curve

    public variable callback = em_nop
    public variable delete_callback_arg = {}

    # Event handling support
    private variable bindtags

    constructor {basewinpath args} {
	# Don't add '.' suffix if there already is one.
	# (This happens chiefly when basewinpath is exactly '.')
	if {![string match {.} \
		[string index $basewinpath \
		[expr [string length $basewinpath]-1]]]} {
	    append basewinpath {.}
	}
	eval $this configure -winpath ${basewinpath}w$this $args

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback em_nop

	# Pack entire subwidget into a subframe.  This makes
	# housekeeping easier, since the only bindings on
	# $winpath will be ones added from inside the class,
	# and also all subwindows can be destroyed by simply
	# destroying $winpath.
	eval frame $winpath $outer_frame_options
	set graph_frame [eval frame $winpath.canvasframe \
		$canvas_frame_options]
	set graph [eval canvas $graph_frame.graph \
		-width $default_width -height $default_height \
		$canvas_options]
	pack $graph -side left -anchor w -fill both -expand 1
	pack $graph_frame -side left -anchor w -fill both -expand 1 
	
	# Initialize color array
	set color [em_array new]
	$color set [list 0 #f00  \
                         1 #0f0  \
	                 2 #00f  \
                         3 #800  \
                         4 #a0a  \
                         5 #0aa  \
                         6 #000  \
                         7 #660  \
                         8 #008  \
                         9 #080   ]
	$color set [list count 10]
 
	# Allocate id and curve arrays
	set id [em_array new]
	set curve [em_array new]

	# Widget bindings
	bind $winpath <Destroy> "$this WinDestroy %W"
	bind $graph <Configure> "$this ResizeCanvas %w %h"

	# Initialize scaling and axes
	$this Rescale

	# Re-enable callbacks
	set callback $callback_keep

        # Event generation support
        set bindtags [list $this $class]
    }
    method ResizeCanvas { width height } { graph } {
	update idletasks  ;# Apparently need this to keep from generating
	## and infinite <Configure> loop
	set border [expr 2*[$graph cget -borderwidth] \
		+2*[$graph cget -highlightthickness]]
	# Note: The <Configure> resize request *includes* both
	#       -borderwidth and -highlightthickness!

	set newwidth [expr $width-$border]
	set newheight [expr $height-$border]
	if {[$graph cget -width]!=$newwidth || \
		[$graph cget -height]!=$newheight} {
	    set confbind [bind $graph <Configure>]
	    bind $graph <Configure> {}  ;# Temporarily disable
	    $graph configure -width $newwidth -height $newheight
	    $this Rescale 0
	    $this DrawCurves
	    bind $graph <Configure> $confbind ;# Reenable
	}
    }
    method SetKeyState { newstate } {
	if {$newstate == $showkey} { return }
	set showkey $newstate
	if {$showkey} {
	    $this DrawKey
	} else {
	    $this HideKey
	}
    }
    method DrawKey { { bbox {} } } {
	# Get bbox from current key, if any, or otherwise
	# use graph extents
	if [string match {} $bbox] {
	    set bbox [$graph bbox key_box]
	}
	set newbox 0 ;# If ==0, then we are overwriting old box
	if [string match {} $bbox] {
	    set bbox [$this GetGraphExtents]
	    set newbox 1
	}

	$this HideKey   ;# Delete any previously existing key
	if { !$showkey } { return }

	# Determine upper righthand corner of new keybox, including
	# adjustment for outer rectangle line width
	set adj [expr ($keywidth+1)/2]
	set key_box_xmax [expr [lindex $bbox 2]-$adj]
	set key_box_ymin [expr [lindex $bbox 1]+$adj]

	# Temporarily write all key text items, in order, anchored
	# on ne corner
	set keycount 0
	set x $key_box_xmax  ; set y $key_box_ymin
	for { set i 0 } { $i < $curvecount } { incr i } {
	    set curveid $i
	    if {[$curve exists $curveid,hidden] && \
		    ![$curve lookup $curveid,hidden] } {
		# Curve exists and is not hidden
		set tempid [$graph create text $x $y \
			-anchor ne \
			-fill [$curve lookup $curveid,color] \
			-justify left \
			-tags {key key_text} \
			-text [$curve lookup $curveid,name]]
		incr keycount
		set bbox [$graph bbox $tempid]
		set y [lindex $bbox 3]
	    }
	}

	# Create appropriately sized key box
	if {$keycount<1} {
	    set keycount 1
	    set x1 $x  ; set y1 $y
	    set x2 $x1 ; set y2 $y1
	} else {
	    set bbox [$graph bbox key_text]
	    set x1 [lindex $bbox 0] ; set y1 [lindex $bbox 1]
	    set x2 [lindex $bbox 2] ; set y2 [lindex $bbox 3]
	}
	set textheight [expr $y2-$y1]
	set textwidth [expr $x2-$x1]
	set lineheight [expr double($textheight)/double($keycount)]
	set x $key_box_xmax  ; set y $key_box_ymin
	if {$newbox} {
	    # Add margin
	    set x [expr $x-$lineheight/4.]
	    set y [expr $y+$lineheight/4.]
	}
	set x1 [expr $x-$textwidth-$lineheight/2.]
	set y1 $y
	set x2 $x
	set y2 [expr $y+$textheight+$lineheight/2.]
	$graph create rectangle $x1 $y1 $x2 $y2 -tag {key key_box}\
		-fill white -outline black -width $keywidth

	# Rewrite all key text items, anchored on nw inside key box
	$graph delete key_text
	set x [expr $x1+$lineheight/4.]
	set y [expr $y1+$lineheight/4.]
	for { set i 0 } { $i < $curvecount } { incr i } {
	    set curveid $i
	    if {[$curve exists $curveid,hidden] && \
		    ![$curve lookup $curveid,hidden] } {
		# Curve exists and is not hidden
		set tempid [$graph create text $x $y \
			-anchor nw \
			-fill [$curve lookup $curveid,color] \
			-justify left \
			-tags {key key_text} \
			-text [$curve lookup $curveid,name]]
		$curve set [list $curveid,keyid $tempid]
		set bbox [$graph bbox $tempid]
		set y [lindex $bbox 3]
	    }
	}

	# Set up motion bindings
	$graph bind key <ButtonPress-1> "$this DragKeyInit %x %y"
	$graph bind key <Button1-Motion> "$this DragKey %x %y"
    }
    method DragKeyInit { x y } { graph but1x but1y } {
	$graph raise key
	set but1x $x
	set but1y $y
    }
    method DragKey { x y } { graph but1x but1y } {
	set delx [expr $x - $but1x]
	set dely [expr $y - $but1y]
	$graph move key $delx $dely
	set but1x $x
	set but1y $y
    }
    method HideKey {} {
	$graph delete key
	$graph bind key <ButtonPress-1> {}
	$graph bind key <Button1-Motion> {}
    }
    method DeleteCurve { curvename } {
        if {![$id exists $curvename]} { return } ;# No such curve
	set curveid [$id lookup $curvename]
	$graph delete tag$curveid
	foreach i [$curve names $curveid,*] {
	    $curve unset $i
	}
	$id unset $curvename
	$this DrawKey      ;# Update key
    }
    method ResetGraph {} {
	# Deletes all data and curves
	foreach name [$id names] {
	    $this DeleteCurve $name
	}
    }
    method ResetData {} {
	# Deletes all data, but leaves curve structures
	foreach name [$id names] {
	    set curveid [$id lookup $name]
	    $curve set [list $curveid,pts {} $curveid,ptsdisp {} \
		    $curveid,lastdisp {}]
	    catch { $curve unset $curveid,xmin }
	    catch { $curve unset $curveid,xmax }
	    catch { $curve unset $curveid,ymin }
	    catch { $curve unset $curveid,xmax }
	    $graph delete tag$curveid
	}
    }
    method NewCurve { curvename } { curvecount id curve color } {
        if {[$id exists $curvename]} {
	    error "Curve $curvename already exists"
	}
	set curveid $curvecount
	incr curvecount
	set curvecolor [$color lookup [expr $curveid % [$color lookup count]]]
	$id set [list $curvename $curveid]
	$curve set [list                     \
		$curveid,name    $curvename  \
		$curveid,color   $curvecolor \
		$curveid,pts      {}         \
		$curveid,ptsdisp  {}         \
		$curveid,lastdisp {}         \
		$curveid,scalevalid 1        \
		$curveid,hidden   0           ]
	catch { $curve unset $curveid,xmin }            ;# Safety
	catch { $curve unset $curveid,xmax }
	catch { $curve unset $curveid,ymin }
	catch { $curve unset $curveid,xmax }

	# Update key
	$this DrawKey

	return $curveid
    }
    method NewCurves { args } {
	foreach curvename $args { $this NewCurve $curvename }
    }
    method GetCurveId { curvename } { id } {
	# Returns empty string if curvename is invalid
	if {[$id exists $curvename]} {
	    return [$id lookup $curvename]
	}
	return
    }
    method ToggleCurveDisplayState { curveid } { curve graph } {
	if {![$curve exists $curveid,hidden]} {
	    error "Curve $curveid doesn't exist"
	}
	set state [$curve lookup $curveid,hidden]
	if {$state} {
	    # Toggle display on (hidden off)
	    $curve set [list $curveid,hidden 0]
	    $this PlotCurve $curveid
	    set color [$curve lookup $curveid,color]
	} else {
	    # Toggle display off (hidden on)
	    $curve set [list $curveid,hidden 1]
	    $graph delete tag$curveid
	    set color [$graph cget -background]
	}
	$this DrawKey      ;# Update key
    }
    method AddPoints { curvename args } { \
	    id curve autodraw autoscale xscale yscale xoff yoff \
	    xgraphmin xgraphmax ygraphmin ygraphmax graph curve_width } {
	set argcount [llength $args]
	if {$argcount<2} { return }
	if {$argcount % 2 != 0} {
	    error "List argument must have even number of elements\n\n\
		    usage:\n\t\$instance AddPoints curvename list\n\n\
		    Invalid call:\n\t$this AddPoints $curvename $args"
	}

	if {[catch {$id lookup $curvename} curveid]} {
	    set curveid [$this NewCurve $curvename]
	}
	if {[catch {$curve getvalues [list $curveid,xmin $curveid,ymin \
		$curveid,xmax $curveid,ymax $curveid,lastdisp]} \
		curve_extents]} {
	    # Initial point 
	    set curve_xmax [set curve_xmin [set x [lindex $args 0]]]
	    set curve_ymax [set curve_ymin [set y [lindex $args 1]]]
	    set xdisp [expr $x*$xscale+$xoff]
	    set ydisp [expr $y*$yscale+$yoff]
	    set xdispstart $xdisp
	    set ydispstart $ydisp
	    $curve set [list  \
		    $curveid,xmin $x   \
		    $curveid,xmax $x   \
		    $curveid,ymin $y   \
		    $curveid,ymax $y   \
		    $curveid,lastdisp \
		     [list [expr $xdisp-100.] [expr $ydisp-100.]] \
		    ]
	} else {
	    set curve_xmin [lindex $curve_extents 0]
	    set curve_ymin [lindex $curve_extents 1]
	    set curve_xmax [lindex $curve_extents 2]
	    set curve_ymax [lindex $curve_extents 3]
	    set drawlast   [lindex $curve_extents 4]
	    set xdispstart [lindex $drawlast 0]
	    set ydispstart [lindex $drawlast 1]
	    unset drawlast
	}

	set xdisplast $xdispstart
	set ydisplast $ydispstart

	set out_of_bounds 0


	set datalist {}
	set drawlist {}
	foreach { x y } $args {
	    ############## BLOCK START
	    ### I had the following block set off as a separate function,
	    ### but the em_class method call overhead in a tight loop is
	    ### a performance killer.
	    append datalist " $x $y"
	    # Data point info
	    if {$x<$curve_xmin} {
		set curve_xmin $x
	    } elseif {$x>$curve_xmax} {
		set curve_xmax $x
	    }
	    if {$y<$curve_ymin} {
		set curve_ymin $y
	    } elseif {$y>$curve_ymax} {
		set curve_ymax $y
	    }
	    # Display point info
	    set xdisp [expr $x*$xscale+$xoff]
	    set ydisp [expr $y*$yscale+$yoff]
	    if {abs($xdisp-$xdisplast)>=1. || abs($ydisp-$ydisplast)>=1.} {
		append drawlist " $xdisp $ydisp"
		set xdisplast $xdisp
		set ydisplast $ydisp
	    }
	    ############## BLOCK END
	    if {$x<$xgraphmin || $x>$xgraphmax || \
		    $y<$ygraphmin || $y>$ygraphmax} {
		set out_of_bounds 1
	    }
	}
	$curve set [list \
		$curveid,xmin $curve_xmin \
		$curveid,xmax $curve_xmax \
		$curveid,ymin $curve_ymin \
		$curveid,ymax $curve_ymax \
		$curveid,lastdisp [list $xdisplast $ydisplast]]
	$curve append $curveid,pts $datalist
	$curve append $curveid,ptsdisp $drawlist

	if {$autodraw && ![$curve lookup $curveid,hidden]} {
	    if {$out_of_bounds && $autoscale} {
		    $this Rescale
		    $this DrawCurves
 	    } elseif {![string match {} $drawlist]} {
		$graph lower [eval $graph create line \
			$xdispstart $ydispstart $drawlist \
			-fill [$curve lookup $curveid,color] \
			-width $curve_width -tags tag$curveid]
	    }
	}
	return $curveid
    }
    method PlotCurve { curveid { startindex 0 }} \
	    { graph curve curve_width showkey } {
	# NOTE 1: If startindex is 0, then tag$curveid is
	#  deleted before the curve is redrawn.
	# NOTE 2: startindex is forced to 0 if "scalevalid"
	#  is false.
	if {![$curve exists $curveid,ptsdisp]} {
	    $graph delete tag$curveid
	    return 0
	}
	if {![$curve lookup $curveid,scalevalid]} {
	    set startindex 0
	    $this RescaleCurve $curveid
	}
	if {$startindex<=0} {
	    $graph delete tag$curveid
	    set startindex 0    ;# Safety
	}
	set ptlist [lrange [$curve lookup $curveid,ptsdisp] \
		[expr 2*$startindex] end]
	set ptcount [expr [llength $ptlist]/2]
	if {$ptcount>=2} {
	    $graph lower [eval $graph create line $ptlist \
		    -fill [$curve lookup $curveid,color] \
		    -width $curve_width -tags tag$curveid]
	}
	return $ptcount	
    }
    method DrawCurves {} { id curve curvecount } {
	# Note: We want to draw curves in order
	set curves_drawn 0
	for { set i 0 } { $i < $curvecount } { incr i } {
	    set curveid $i
	    if {[$curve exists $curveid,hidden] && \
		    ![$curve lookup $curveid,hidden] } {
		# Curve exists and is not hidden
		$this PlotCurve $curveid
		incr curves_drawn
	    }
	}
	return $curves_drawn
    }
    method SetXLabel { newxlabel } { xlabel } {
	set xlabel $newxlabel
    }
    method SetYLabel { newylabel } { ylabel } {
	# Separate ylabel characters by newlines
	set ylabel [join [split $newylabel {}] "\n"]
    }
    method SetLabels { newxlabel newylabel } {} {
	$this SetXLabel $newxlabel
	$this SetYLabel $newylabel
	$this DrawAxes
    }
    method GetLabels {} { xlabel ylabel } {
	return [list $xlabel $ylabel]
    }
    method GetGraphExtents {} {
	set xdispmin $lmargin
	set xdispmax [expr $xgraphmax*$xscale+$xoff]
	set ydispmin $tmargin
	set ydispmax [expr $ygraphmin*$yscale+$yoff]
	return [list $xdispmin $ydispmin $xdispmax $ydispmax]
    }
    method DrawAxes {} {
	$graph delete axes
	set canwidth [$graph cget -width]
	set canheight [$graph cget -height]
	set graphextents [$this GetGraphExtents]
	set xdispmin [lindex $graphextents 0]
	set ydispmin [lindex $graphextents 1]
	set xdispmax [lindex $graphextents 2]
	set ydispmax [lindex $graphextents 3]
	$graph create line \
		$xdispmin $ydispmin \
		$xdispmin $ydispmax \
		$xdispmax $ydispmax \
		$xdispmax $ydispmin \
		$xdispmin $ydispmin \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create text $xdispmin [expr $ydispmax+$bmargin*0.1] \
		-text $xgraphmin -anchor n -tags axes
	$graph create text $xdispmax [expr $ydispmax+$bmargin*0.1] \
		-text $xgraphmax -anchor n -tags axes
	$graph create text [expr $xdispmin-$lmargin*0.1] $ydispmax \
		-text $ygraphmin -anchor se -tags axes
	$graph create text [expr $xdispmin-$lmargin*0.1] $ydispmin \
		-text $ygraphmax -anchor ne -tags axes

	# Place labels
	$graph create text \
		[expr ($xdispmin+$xdispmax)/2.] \
		[expr $canheight-$bmargin*0.1]  \
		-text $xlabel -anchor s -tags axes
	$graph create text \
		[expr $lmargin*0.2]             \
		[expr ($ydispmin+$ydispmax)/2.] \
		-text $ylabel -anchor w -justify center -tags axes
	
	# X-axes tics
	set xquart [expr ($xdispmax-$xdispmin)/4.]
	set x1 [expr $xdispmin+$xquart]
	set y1 $ydispmax  ;  set y2 [expr $y1-1.4*$axeswidth]
	set y3 $ydispmin  ;  set y4 [expr $y3+1.4*$axeswidth]
	$graph create line $x1 $y1 $x1 $y2 \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create line $x1 $y3 $x1 $y4 \
		-fill $axescolor -width $axeswidth -tags axes
	set x1 [expr $xdispmax-$xquart]
	$graph create line $x1 $y1 $x1 $y2 \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create line $x1 $y3 $x1 $y4 \
		-fill $axescolor -width $axeswidth -tags axes
	set x1 [expr $xdispmin+2*$xquart]
	set y1 $ydispmax  ;  set y2 [expr $y1-2*$axeswidth]
	set y3 $ydispmin  ;  set y4 [expr $y3+2*$axeswidth]
	$graph create line $x1 $y1 $x1 $y2 \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create line $x1 $y3 $x1 $y4 \
		-fill $axescolor -width $axeswidth -tags axes
	# Y-axes tics
	set yquart [expr ($ydispmax-$ydispmin)/4.]
	set y1 [expr $ydispmax-$yquart]
	set x1 $xdispmin  ;  set x2 [expr $x1+1.4*$axeswidth]
	set x3 $xdispmax  ;  set x4 [expr $x3-1.4*$axeswidth]
	$graph create line $x1 $y1 $x2 $y1 \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create line $x3 $y1 $x4 $y1 \
		-fill $axescolor -width $axeswidth -tags axes
	set y1 [expr $ydispmin+$yquart]
	$graph create line $x1 $y1 $x2 $y1 \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create line $x3 $y1 $x4 $y1 \
		-fill $axescolor -width $axeswidth -tags axes
	set y1 [expr $ydispmax-2*$yquart]
	set x1 $xdispmin  ;  set x2 [expr $x1+2*$axeswidth]
	set x3 $xdispmax  ;  set x4 [expr $x3-2*$axeswidth]
	$graph create line $x1 $y1 $x2 $y1 \
		-fill $axescolor -width $axeswidth -tags axes
	$graph create line $x3 $y1 $x4 $y1 \
		-fill $axescolor -width $axeswidth -tags axes
    }
    private method RescaleCurve { curveid } \
	    { curve xscale yscale xoff yoff } {
	# Select and rescale display points
	# NOTE: This routine may change the number of points
	#  in the ptsdisp list.
	if {![$curve exists $curveid,pts]} { return }
	set ptslist [$curve lookup $curveid,pts]
	set xprev [lindex $ptslist 0]
	set yprev [lindex $ptslist 1]
	if [string match {} $yprev] {
	    # No points
	    $curve set [list $curveid,ptsdisp {} \
		    $curveid,lastdisp {} \
		    $curveid,scalevalid 1]
	    return
	}
	set xdisp [expr $xprev*$xscale+$xoff]
	set ydisp [expr $yprev*$yscale+$yoff]
	set newlist "$xdisp $ydisp"
	# Add all points that will display at distinct screen locations
	set xdiff 1.  ; set ydiff 1.   ;# Safety init
	if {$xscale!=0.} {set xdiff [expr 1./abs($xscale)] }
	if {$yscale!=0.} {set ydiff [expr 1./abs($yscale)] }
	foreach { x y } [lrange $ptslist 2 end] {
	    if { abs($x-$xprev)<$xdiff && abs($y-$yprev)<$ydiff } {
		continue
	    }
	    set xprev $x ; set yprev $y
	    set xdisp [expr $x*$xscale+$xoff]
	    set ydisp [expr $y*$yscale+$yoff]
	    append newlist " $xdisp $ydisp"
	}
	$curve set [list $curveid,ptsdisp $newlist \
		$curveid,lastdisp [list $xdisp $ydisp] \
		$curveid,scalevalid 1]
    }
    method Rescale { {use_data 1} } {
	# Set import use_data to non-zero to invoke data-based
	# graph extents changes.  Either way, rescales the
	# graph as necessary to fill the canvas.
	if {$use_data} {
	    # Find data extents
	    for { set cid 0 } { $cid < $curvecount } { incr cid } {
		if {![$curve exists $cid,xmin] || \
			[$curve lookup $cid,hidden] } {
		    continue     ;# Empty or hidden curve
		}
		if {![info exists xmin]} {
		    # First non-empty curve
		    set xmin [$curve lookup $cid,xmin]
		    set xmax [$curve lookup $cid,xmax]
		    set ymin [$curve lookup $cid,ymin]
		    set ymax [$curve lookup $cid,ymax]
		} else {
		    # Normal case
		    set test [$curve lookup $cid,xmin]
		    if { $test < $xmin } { set xmin $test }
		    set test [$curve lookup $cid,xmax]
		    if { $test > $xmax } { set xmax $test }
		    set test [$curve lookup $cid,ymin]
		    if { $test < $ymin } { set ymin $test }
		    set test [$curve lookup $cid,ymax]
		    if { $test > $ymax } { set ymax $test }
		}
	    }
	    if {![info exists xmin]} { 
		# No data, so fill in dummy values
		set xmin 0  ; set ymin 0
		set xmax 1  ; set ymax 1
	    }
	    if { $xmin >= $xmax } { set xmax [expr $xmin +1] } ;# Safety
	    if { $ymin >= $ymax } { set ymax [expr $ymin +1] }

	    # Adjust to "round" numbers
	    set xdelta [expr $xmax-$xmin]
	    if {$xdelta<=0.0} {set xdelta 1. } ;# Safety 
	    set xrangescale [expr pow(10,ceil(-log10($xdelta)))]
	    set xmin [expr (floor($xmin*$xrangescale))/$xrangescale]
	    set xmax [expr (ceil($xmax*$xrangescale))/$xrangescale]
	    set ydelta [expr $ymax-$ymin]
	    if {$ydelta<=0.0} {set ydelta 1. } ;# Safety 
	    set yrangescale [expr pow(10,ceil(-log10($ydelta)))]
	    set ymin [expr (floor($ymin*$yrangescale))/$yrangescale]
	    set ymax [expr (ceil($ymax*$yrangescale))/$yrangescale]

	    # Round and save for future use
	    set xgraphmin [format "%g" $xmin]
	    set xgraphmax [format "%g" $xmax]
	    if { $xgraphmax<=$xgraphmin } {
		# Too much rounding.  Try again
		set xgraphmin [format "%.15g" $xmin]
		set xgraphmax [format "%.15g" $xmax]
	    }
	    set ygraphmin [format "%g" $ymin]
	    set ygraphmax [format "%g" $ymax]
	    if { $ygraphmax<=$ygraphmin } {
		# Too much rounding.  Try again
		set ygraphmin [format "%.15g" $ymin]
		set ygraphmax [format "%.15g" $ymax]
	    }
	} ;# End of if {$use_data}

 	if {$xgraphmax<=$xgraphmin} {
	    # Safety.  This hopefully won't happen
	    set xgraphmax [expr $xgraphmin + 1. ]
	}
 	if {$ygraphmax<=$ygraphmin} {
	    # Safety.  This hopefully won't happen
	    set ygraphmax [expr $ygraphmin + 1. ]
	}

	# Calculate lower bound for margin size from axes coordinate text
	$graph delete margin_test   ;# Safety
	$graph create text 0 0 -text $xgraphmin -anchor sw -tags margin_test
	$graph create text 0 0 -text $xgraphmax -anchor sw -tags margin_test
	set bbox [$graph bbox margin_test]
	set xtext_height [expr ([lindex $bbox 3]-[lindex $bbox 1])*1.2+1]
	set xtext_halfwidth [expr ([lindex $bbox 2]-[lindex $bbox 0])*0.6+1]
	## Include a little bit extra
	$graph delete margin_test
	$graph create text 0 0 -text $ygraphmin -anchor sw -tags margin_test
	$graph create text 0 0 -text $ygraphmax -anchor sw -tags margin_test
	set bbox [$graph bbox margin_test]
	set ytext_halfheight [expr ([lindex $bbox 3]-[lindex $bbox 1])*0.6+1]
	set ytext_width [expr ([lindex $bbox 2]-[lindex $bbox 0])*1.2+1]
	## Include a little bit extra
	$graph delete margin_test
	## Increase left and top text margin if necessary
	if {$xtext_halfwidth>$ytext_width} {
	    set ytext_width $xtext_halfwidth
	}
	if {$ytext_halfheight>$xtext_height} {
	    set xtext_height $ytext_halfheight
	}

	# Set origin and scaling factors
	# NOTE: Canvas widget has +y downward
	set canwidth [$graph cget -width]
	set canheight [$graph cget -height]
	set xorg_new $xgraphmin
	set yorg_new $ygraphmax
	set lmargin_new [expr $canwidth*(1-$graphsize)/2.]
	set rmargin_new $lmargin_new
	set tmargin_new [expr $canheight*(1-$graphsize)/2.]
	set bmargin_new $tmargin_new
	if { $lmargin_new < $ytext_width }  {
	    set lmargin_new $ytext_width
	}
	if { $rmargin_new < $xtext_halfwidth } {
	    set rmargin_new $xtext_halfwidth
	}
	if { $bmargin_new < $xtext_height }  {
	    set bmargin_new $xtext_height
	}
	if { $tmargin_new < $ytext_halfheight } {
	    set tmargin_new $ytext_halfheight
	}
	set xscale_new  [expr ($canwidth-$lmargin_new-$rmargin_new) \
		/double($xgraphmax-$xgraphmin)]
	set yscale_new  [expr ($canheight-$tmargin_new-$bmargin_new) \
		/double($ygraphmin-$ygraphmax)]

	# See if any changes were made
	if {[info exists xorg] && $xorg_new==$xorg && $yorg_new==$yorg && \
		$lmargin_new==$lmargin && $rmargin_new==$rmargin && \
		$tmargin_new==$tmargin && $bmargin_new==$bmargin && \
		$xscale_new==$xscale && $yscale_new==$yscale} {
	    # No changes, so no need to rescale data or redraw axes
	    return
	}
	set xorg    $xorg_new        ; set yorg    $yorg_new
	set lmargin $lmargin_new     ; set rmargin $rmargin_new
	set tmargin $tmargin_new     ; set bmargin $bmargin_new
	set xscale  $xscale_new      ; set yscale  $yscale_new

	# Auxillary transformation variables
	set xoff [expr $lmargin-$xorg*$xscale]
	set yoff [expr $tmargin-$yorg*$yscale]

	# Rescale display points on all non-hidden curves
	for { set cid 0 } { $cid < $curvecount } { incr cid } {
	    if {[$curve exists $cid,pts]} {
		if {[$curve lookup $cid,hidden]} {
		    $curve set [list $cid,scalevalid 0]
		    ## Mark scaling invalid
		} else {
		    $this RescaleCurve $cid
		}
	    }
	}
	# Delete current graph and redraw
	set bbox [$graph bbox key_box] ;# Save current box location
	$graph delete all
	$this DrawAxes
	$this DrawCurves
	$this DrawKey $bbox
    }
    method PrintData { curvename } {
	# For Debugging
	if {![$id exists $curvename]} {
	    # No match
	    return
	}
	set curveid [$id lookup $curvename]
	foreach i [lsort -ascii [$curve names $curveid,*]] {
	    puts "$i = [$curve lookup $i]"
	}
	return
    }
    method HasGui {} {
        return 0
    }
    method WinDestroy { w } { winpath } {
	if {[string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this delete
    }
    destructor {
	# Notify client of impending destruction
	eval $callback $this DELETE {$delete_callback_arg}
        em_eventHandler Generate delete $bindtags

	catch { $color delete }
	catch { $id delete }
	catch { $curve delete }
	if {[info exists winpath] && [winfo exists $winpath]} {
            # Remove <Destroy> binding
	    bind $winpath <Destroy> {}
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}


