# FILE: selectbox.tcl
#
# Selection box dialog sub-widget
#
#   This subwidget provides a wrapper around a 2-entry multiwidget,
# (Path: and File:), allowing path+filename to be treated as a single
# entity.  The -callback function is invoked whenever a callback is
# obtained from the multiwidget (check the multiwidget docs, but
# currently this occurs only on a <Key-Return> event in one of the entry
# boxes).  On this event, the current _displayed_ values are read from
# the Path: and File: entry boxes, combined into one complete pathname
# and stored in the variable -value.  Then -callback is called with
# ACTIONID set to SELECT, with args "-value $value".  The client routine
# may at any time access -value to get the last committed value, or may
# call ReadFilename to get the value as currently displayed.
#   The routine SetFullname is provided to fill the Path: and File:
# entries from without.  This sets the committed value as well as the
# displayed text.  Use SetPath and SetFilename to set this fields
# individually.
#   The default title bar value is "Selection".  This can be set via
# the -title option at construct time.  Set to {} to turn the title
# off.  Other options include -labelwidth and -valuewidth to set the
# sizes of the labels and text entry areas, respectively.
#
# debugging enable/disable/level

em_class omf_selectbox {
    const public variable winpath
    const public variable outer_frame_options = "-bd 0 -relief flat"
    const private variable multibox
    public variable value = {}
    const public variable title = "Selection"
    const public variable labelwidth = 5
    const public variable valuewidth = 16
    public variable callback = em_nop
    const private variable pathlabel = "Path:"
    const private variable filelabel = "File:"
    constructor {basewinpath args} {
	# Don't add '.' suffix if there already is one.
	# (This happens chiefly when basewinpath is exactly '.')
	if {![string match {.} \
		[string index $basewinpath \
		[expr [string length $basewinpath]-1]]]} {
	    append basewinpath {.}
	}
	eval $this configure -winpath ${basewinpath}w$this $args
	eval frame $winpath $outer_frame_options
	## Note: We need a separate outer frame because we can't
	##  bind to <Destroy> on the multibox widget (as usual).

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback em_nop

	# Create multibox
	set opts [list -labellist "$pathlabel {} $filelabel {}"]
	lappend opts -callback "$this MultiboxCallback"
	if {![string match {} $labelwidth]} {
	    lappend opts -labelwidth $labelwidth
	}
	if {![string match {} $valuewidth]} {
	    lappend opts -valuewidth $valuewidth
	}
	if {![string match {} $title]} {
	    lappend opts -title $title
	}
	set multibox [eval omf_multibox new {$winpath} $opts]
	$multibox EntryConfigure $pathlabel -display_preference "end"
	pack [$multibox cget -winpath] -fill x -expand 0
	
	bind $winpath <Destroy> "$this WinDestroy %W"

	# Re-enable callbacks
	set callback $callback_keep
    }
    method ReadFilename {} { multibox pathlabel filelabel } {
	set pathvalue [$multibox LookUp $pathlabel]
	set filevalue [$multibox LookUp $filelabel]
	set fullname [file join $pathvalue $filevalue]
    }
    method MultiboxCallback { subwidget actionid args } { value callback } {
	set value [$this ReadFilename]
	eval $callback $this SELECT -value {$value}
    }
    method SetFullname { fullname } { value pathlabel filelabel multibox } {
	set value $fullname
	set pathvalue [file dirname $fullname]
	set filevalue [file tail $fullname]
	$multibox Set $pathlabel $pathvalue $filelabel $filevalue
    }
    method SetFilename { filevalue } {value multibox filelabel} {
	set pathvalue [file dirname $value]
	set value [file join $pathvalue $filevalue]
	$multibox Set $filelabel $filevalue
    }
    method SetPath { pathvalue } {value multibox pathlabel} {
	set filevalue [file tail $value]
	set value [file join $pathvalue $filevalue]
	$multibox Set $pathlabel $pathvalue
    }
    method WinDestroy { w } { winpath } {
	if {[info exists winpath] && [string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this delete
    }
    destructor {
	if {[info exists winpath] && [winfo exists $winpath]} {
            # Remove <Destroy> binding
	    bind $winpath <Destroy> {}
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}
